/**
 * Author = Robbert Harms
 * Date = 2016-02-09
 * License = LGPL v3
 * Maintainer = Robbert Harms
 * Email = robbert.harms@maastrichtuniversity.nl
 */

double bessel_j0(double x);
double bessel_i0(double x);
double log_bessel_i0(double x);
